@@ -1,27 +1,11 @@
-/*
-  Racket
-  Copyright (c) 2004-2016 PLT Design Inc.
-
-  This library is free software; you can redistribute it and/or
-  modify it under the terms of the GNU Library General Public
-  License as published by the Free Software Foundation; either
-  version 2 of the License, or (at your option) any later version.
-
-  This library is distributed in the hope that it will be useful,
-  but WITHOUT ANY WARRANTY; without even the implied warranty of
-  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
-  Library General Public License for more details.
-
-  You should have received a copy of the GNU Library General Public
-  License along with this library; if not, write to the Free
-  Software Foundation, Inc., 51 Franklin Street, Fifth Floor,
-  Boston, MA 02110-1301 USA.
-
-*/
-
 #include "schpriv.h"
+#include "schmach.h"
 
-/* PLAN:
+/* This file implements the compiler's letrec-check pass.
+ *
+ * See "eval.c" for an overview of compilation passes.
+ * 
+ * PLAN:
  *
  * Imagine starting with a simple abstract interpretation: traverse
  * the program in evaluation order, treating `if` like `begin` and
@@ -109,7 +93,6 @@ void scheme_init_letrec_check()
 #define LET_CHECKED   (0x1 << 2)
 
 #define FRAME_TYPE_LETREC   1
-#define FRAME_TYPE_LETSTAR  2
 #define FRAME_TYPE_LET      3
 #define FRAME_TYPE_CLOSURE  4
 #define FRAME_TYPE_TOP      5
@@ -140,9 +123,6 @@ typedef struct Letrec_Check_Frame {
      been wrapped with an #<unsafe-undefined> check */
   int *ref;
 
-  /* so we can get to variable names */
-  Scheme_Let_Header *head;
-
   /* we keep a list of all deferred expressions, only so that we can
      drop the body for any that are not processed (which means that
      they won't be used) */
@@ -160,8 +140,9 @@ typedef struct Scheme_Deferred_Expr {
      only need to process it once */
   int done;
 
-  /* the expression that has been deferred */
-  Scheme_Closure_Data *expr;
+  /* the expression that has been deferred -- usally a Scheme_Lambda,
+     but can be a Scheme_IR_Local for propagation */
+  Scheme_Object *expr;
 
   /* the frame that existed when the expr was deferred */
   Letrec_Check_Frame *frame;
@@ -178,12 +159,13 @@ static Letrec_Check_Frame *init_letrec_check_frame(int frame_type, int subexpr,
                                                    mzshort count,
                                                    Letrec_Check_Frame *prev,
                                                    Letrec_Check_Frame *share_with,
-                                                   Scheme_Let_Header *head)
+                                                   Scheme_IR_Let_Header *head,
+                                                   Scheme_Lambda *lam)
 {
   Scheme_Deferred_Expr **chain;
   Letrec_Check_Frame *frame;
   Scheme_Object **def;
-  int *ref, i;
+  int *ref, i, j, pos;
 
   frame = (Letrec_Check_Frame *)MALLOC_ONE_RT(Letrec_Check_Frame);
 #ifdef MZTAG_REQUIRED
@@ -195,7 +177,23 @@ static Letrec_Check_Frame *init_letrec_check_frame(int frame_type, int subexpr,
   frame->count = count;
   frame->next = prev;
 
-  frame->head = head;
+  if (head) {
+    Scheme_IR_Let_Value *irlv = (Scheme_IR_Let_Value *)head->body;
+    pos = 0;
+    for (i = head->num_clauses; i--; irlv = (Scheme_IR_Let_Value *)irlv->body) {
+      for (j = 0; j < irlv->count; j++) {
+        irlv->vars[j]->mode = SCHEME_VAR_MODE_LETREC_CHECK;
+        irlv->vars[j]->letrec_check.frame = frame;
+        irlv->vars[j]->letrec_check.frame_pos = pos++;
+      }
+    }
+  } else if (lam) {
+    for (j = lam->num_params; j--; ) {
+      lam->ir_info->vars[j]->mode = SCHEME_VAR_MODE_LETREC_CHECK;
+      lam->ir_info->vars[j]->letrec_check.frame = frame;
+      lam->ir_info->vars[j]->letrec_check.frame_pos = j;
+    }
+  }
 
   if (share_with) {
     /* Moving from RHS phase to BODY phase for `let[rec]`,
@@ -256,33 +254,10 @@ static Letrec_Check_Frame *get_nearest_rhs(Letrec_Check_Frame *frame)
 
 /* returns the frame that was created when pos was created, and
    changes pos to be relative to that frame */
-static Letrec_Check_Frame *get_relative_frame(int *pos, Letrec_Check_Frame *frame)
+static Letrec_Check_Frame *get_relative_frame(int *pos, Scheme_IR_Local *var)
 {
-  while (1) {
-    /* we've gone wrong if pos_int is negative or if the frame has
-       become NULL because pos should have be a valid LHS variable
-       reference */
-    SCHEME_ASSERT(*pos >= 0, "get_relative_frame: pos is negative");
-    SCHEME_ASSERT(frame, "get_relative_frame: frame is NULL");
-
-    /* if we're in the RHS of a let, no bindings for the LHS variables
-       have been pushed yet, pos can't possibly be in this frame.  so
-       don't do any offsetting and look in the next frame */
-    if ((frame->frame_type == FRAME_TYPE_LET)
-        && (frame->subexpr == LET_RHS_EXPR)) {
-      frame = frame->next;
-      /* recur */
-    } else {
-      if (*pos >= frame->count) {
-        /* we're not in the right frame yet, so offset pos by the
-           number of bindings in this frame */
-        (*pos) -= frame->count;
-        frame = frame->next;
-        /* recur */
-      } else
-        return frame;
-    }
-  }
+  *pos = var->letrec_check.frame_pos;
+  return var->letrec_check.frame;
 }
 
 /* adds expr to the deferred bindings of lhs */
@@ -293,6 +268,9 @@ static void update_frame(Letrec_Check_Frame *outer, Letrec_Check_Frame *inner,
 
   SCHEME_ASSERT(position < outer->count, "update_frame: position exceeds binding count");
 
+  if (SCHEME_VOIDP(outer->def[position]))
+    scheme_signal_error("oops\n");
+    
   /* put the deferred expression in the right place */
   prev_def = outer->def[position];
   prev_def = scheme_make_pair((Scheme_Object *)clos, prev_def); 
@@ -303,7 +281,7 @@ static void update_frame(Letrec_Check_Frame *outer, Letrec_Check_Frame *inner,
 }
 
 /* creates a deferred expression "closure" by closing over the frame */
-static Scheme_Deferred_Expr *make_deferred_expr_closure(Scheme_Closure_Data *expr, Letrec_Check_Frame *frame)
+static Scheme_Deferred_Expr *make_deferred_expr_closure(Scheme_Object *expr, Letrec_Check_Frame *frame)
 {
   Scheme_Deferred_Expr *clos;
 
@@ -321,11 +299,11 @@ static Scheme_Deferred_Expr *make_deferred_expr_closure(Scheme_Closure_Data *exp
 
 static Scheme_Object *letrec_check_expr(Scheme_Object *, Letrec_Check_Frame *, Scheme_Object *);
 
-static void letrec_check_lets_resume(Letrec_Check_Frame *frame, Scheme_Let_Header *head)
+static void letrec_check_lets_resume(Letrec_Check_Frame *frame, Scheme_IR_Let_Header *head)
 {
-  Scheme_Compiled_Let_Value *clv;
+  Scheme_IR_Let_Value *irlv;
   Scheme_Object *body;
-  int i, j, k, *clv_flags;
+  int i, j, k;
   int was_checked;
 
   body = head->body;
@@ -334,76 +312,48 @@ static void letrec_check_lets_resume(Letrec_Check_Frame *frame, Scheme_Let_Heade
        that we have invalidated; i.e., adding check-undefineds around
        references means there is one (more) instance where the LHS
        variable is not used in application position */
-    k = head->count;
+    k = 0;
     for (i = head->num_clauses; i--;) {
-      clv = (Scheme_Compiled_Let_Value *) body;
-      clv_flags = clv->flags;
-      k -= clv->count;
-      for (j = 0; j < clv->count; j++) {
+      irlv = (Scheme_IR_Let_Value *) body;
+      for (j = 0; j < irlv->count; j++) {
         was_checked = (frame->ref[k + j] & LET_CHECKED);
-        if (was_checked) {
-          clv_flags[j] -= (clv_flags[j] & SCHEME_WAS_ONLY_APPLIED);
-          clv_flags[j] -= (clv_flags[j] & SCHEME_WAS_APPLIED_EXCEPT_ONCE);
-        } 
+        if (was_checked)
+          irlv->vars[j]->non_app_count = irlv->vars[j]->use_count;
       }
-      clv->flags = clv_flags;
-      clv->names = NULL; /* not used in later passes */
-      body = clv->body;
+      k += irlv->count;
+      body = irlv->body;
     }
   }
 }
 
 /* records that we have seen a reference to loc */
-static Scheme_Object *record_checked(Scheme_Local *loc, Letrec_Check_Frame *frame)
+static Scheme_Object *record_checked(Scheme_IR_Local *loc, Letrec_Check_Frame *frame)
 {
-  int position = SCHEME_LOCAL_POS(loc), k;
-  Scheme_Object *obj;
-
-  frame = get_relative_frame(&position, frame);
+  int position;
+  
+  frame = get_relative_frame(&position, loc);
   frame->ref[position] |= LET_CHECKED;
 
-  obj = frame->head->body;
-  k = frame->head->count;
-
-  while (1) {
-    Scheme_Compiled_Let_Value *clv = (Scheme_Compiled_Let_Value *)obj;
-      
-    SCHEME_ASSERT(SAME_TYPE(SCHEME_TYPE(obj), scheme_compiled_let_value_type), "not a clv");
-    SCHEME_ASSERT(position >= 0, "position went negative");
-      
-    k -= clv->count;
-
-    if (position >= k)
-      return clv->names[position - k];
-      
-    obj = clv->body;
-  }
-
-  ESCAPED_BEFORE_HERE;
+  return loc->name;
 }
 
 static Scheme_Object *letrec_check_local(Scheme_Object *o, Letrec_Check_Frame *frame,
                                          Scheme_Object *pos)
 {
   Letrec_Check_Frame *in_frame;
-  Scheme_Local *loc = (Scheme_Local *)o;
-  int position, in_position;
+  Scheme_IR_Local *loc = (Scheme_IR_Local *)o;
+  int in_position;
 
-  position = SCHEME_LOCAL_POS(loc);
-
-  in_position = position;
-  in_frame = get_relative_frame(&in_position, frame);
+  in_frame = get_relative_frame(&in_position, loc);
 
   if (SCHEME_FALSEP(pos)) {
     /* mark as potentially applied (i.e., in an "unsafe" context)
-       for deferred closures (gre "EXPL-4" for information): */
+       for deferred closures (grep "EXPL-4" for information): */
     if (in_frame->ref)
       in_frame->ref[in_position] |= LET_APPLY_USE;
   } else {
     /* propagate any deferred expressions (grep "EXPL-2" for information): */
-    if (in_frame->def
-        && !SCHEME_NULLP(in_frame->def[in_position])
-        && !SCHEME_NULLP(pos)) {
+    if (!SCHEME_NULLP(pos)) {
       Letrec_Check_Frame *outer_frame;
       Scheme_Object *ls;
       outer_frame = get_nearest_rhs(frame);
@@ -418,9 +368,16 @@ static Scheme_Object *letrec_check_local(Scheme_Object *o, Letrec_Check_Frame *f
           pos = SCHEME_CDR(pos);
         }
 
-        ls = scheme_append(in_frame->def[in_position],
-                           outer_frame->def[dpos]);
+        ls = (Scheme_Object *)make_deferred_expr_closure(o, frame);
+        ls = scheme_make_pair(ls, outer_frame->def[dpos]);
         outer_frame->def[dpos] = ls;
+
+        /* But if we're trying to propagate to a variable that was
+           already referenced, we need to treat this one as
+           referenced, too */
+        if ((outer_frame->ref[dpos] & LET_APPLY_USE)
+            && in_frame->ref)
+          in_frame->ref[in_position] |= LET_APPLY_USE;
       }
     }
   }
@@ -454,10 +411,9 @@ static Scheme_Object *letrec_check_local(Scheme_Object *o, Letrec_Check_Frame *f
   return o;
 }
 
-static int is_effect_free_prim(Scheme_Object *rator)
+static int is_effect_free_prim(Scheme_Object *rator, int argc)
 {
-  if (SCHEME_PRIMP(rator)
-      && (SCHEME_PRIM_PROC_OPT_FLAGS(rator) & SCHEME_PRIM_IS_OMITABLE_ANY))
+  if (SCHEME_PRIMP(rator) && scheme_is_omitable_primitive(rator, argc))
     return 1;
 
   return 0;
@@ -474,7 +430,7 @@ static Scheme_Object *letrec_check_application(Scheme_Object *o, Letrec_Check_Fr
   /* we'll have to check the rator and all the arguments */
   n = 1 + app->num_args;
 
-  if (is_effect_free_prim(app->args[0])) {
+  if (is_effect_free_prim(app->args[0], app->num_args)) {
     /* an immediate prim cannot call anything among its arguments */
   } else {
     /* argument might get applied */
@@ -496,7 +452,7 @@ static Scheme_Object *letrec_check_application2(Scheme_Object *o, Letrec_Check_F
 
   app = (Scheme_App2_Rec *)o;
     
-  if (is_effect_free_prim(app->rator)) {
+  if (is_effect_free_prim(app->rator, 1)) {
     /* an immediate prim cannot call anything among its arguments */
   } else {
     /* argument might get applied */
@@ -518,7 +474,7 @@ static Scheme_Object *letrec_check_application3(Scheme_Object *o, Letrec_Check_F
 
   app = (Scheme_App3_Rec *)o;
 
-  if (is_effect_free_prim(app->rator)) {
+  if (is_effect_free_prim(app->rator, 2)) {
     /* an immediate prim cannot call anything among its arguments */
   } else {
     /* argument might get applied */
@@ -590,56 +546,54 @@ static Scheme_Object *letrec_check_wcm(Scheme_Object *o, Letrec_Check_Frame *fra
   return o;
 }
 
-static Scheme_Object *letrec_check_closure_compilation(Scheme_Object *o, Letrec_Check_Frame *frame, Scheme_Object *pos)
+static Scheme_Object *letrec_check_lambda(Scheme_Object *o, Letrec_Check_Frame *frame, Scheme_Object *pos)
 {
-  Scheme_Closure_Data *data;
+  Scheme_Lambda *lam;
   Letrec_Check_Frame *new_frame;
   Scheme_Object *val;
   int num_params;
 
-  data = (Scheme_Closure_Data *)o;
+  lam = (Scheme_Lambda *)o;
 
   /* if we have not entered a letrec, pos will be false */
   if (SCHEME_FALSEP(pos)) {
-    num_params = data->num_params;
+    num_params = lam->num_params;
     new_frame = init_letrec_check_frame(FRAME_TYPE_CLOSURE, LET_BODY_EXPR,
                                         num_params, frame, NULL,
-                                        NULL);
+                                        NULL, lam);
 
     SCHEME_ASSERT(num_params >= 0, "lambda has negative arguments what do");
         
-    val = letrec_check_expr(data->code, new_frame, pos);
-    data->code = val;
+    val = letrec_check_expr(lam->body, new_frame, pos);
+    lam->body = val;
   } else {
     /* we can defer this lambda because it is not inside an
        application! hurray! */
     Scheme_Deferred_Expr *clos;
     Letrec_Check_Frame *outer_frame = NULL;
 
-    if (!SCHEME_NULLP(pos)) {
-      /* pos is either a single integer or a list of integers */
+    /* pos is either a single integer or a list of integers */
 
-      /* create a deferred expression that closes over the frame it
-         appeared in, and update the frame where the binding lives
-         (which may be an enclosing frame) */
-      outer_frame = get_nearest_rhs(frame);
-      clos = make_deferred_expr_closure(data, frame);
+    /* create a deferred expression that closes over the frame it
+       appeared in, and update the frame where the binding lives
+       (which may be an enclosing frame) */
+    outer_frame = get_nearest_rhs(frame);
+    clos = make_deferred_expr_closure((Scheme_Object *)lam, frame);
 
-      while (SCHEME_INTP(pos) || SCHEME_PAIRP(pos)) {
-        int position;
+    while (SCHEME_INTP(pos) || SCHEME_PAIRP(pos)) {
+      int position;
 
-        if (SCHEME_INTP(pos)) {
-          position = SCHEME_INT_VAL(pos);
-          pos = scheme_null;
-        } else {
-          position = SCHEME_INT_VAL(SCHEME_CAR(pos));
-          pos = SCHEME_CDR(pos);
-        }
-
-        /* attach the deferred_expr_closure to the right position 
-           in the correct frame */
-        update_frame(outer_frame, frame, position, clos);
+      if (SCHEME_INTP(pos)) {
+        position = SCHEME_INT_VAL(pos);
+        pos = scheme_null;
+      } else {
+        position = SCHEME_INT_VAL(SCHEME_CAR(pos));
+        pos = SCHEME_CDR(pos);
       }
+
+      /* attach the deferred_expr_closure to the right position 
+         in the correct frame */
+      update_frame(outer_frame, frame, position, clos);
     }
   }
 
@@ -649,12 +603,12 @@ static Scheme_Object *letrec_check_closure_compilation(Scheme_Object *o, Letrec_
 static void letrec_check_deferred_expr(Scheme_Object *o)
 {
   Scheme_Deferred_Expr *clos;
-  Scheme_Closure_Data *data;
+  Scheme_Lambda *lam;
   Letrec_Check_Frame *inner, *new_frame;
   Scheme_Object *val;
   int num_params;
 
-  /* gets the closed over data from clos, which will always be a
+  /* gets the closed over lam from clos, which will always be a
      deferred expression that contains a closure */
   clos = (Scheme_Deferred_Expr *)o;
 
@@ -665,25 +619,27 @@ static void letrec_check_deferred_expr(Scheme_Object *o)
   SCHEME_ASSERT(SAME_TYPE(SCHEME_TYPE(clos), scheme_deferred_expr_type),
                 "letrec_check_deferred_expr: clos is not a scheme_deferred_expr");
 
-  data = (Scheme_Closure_Data *)clos->expr;
-  SCHEME_ASSERT(SAME_TYPE(SCHEME_TYPE(data), scheme_compiled_unclosed_procedure_type),
-                "deferred expression does not contain a lambda");
+  if (SAME_TYPE(SCHEME_TYPE(clos->expr), scheme_ir_lambda_type)) {
+    lam = (Scheme_Lambda *)clos->expr;
 
-  inner = clos->frame;
-
-  num_params = data->num_params;
-
-  new_frame = init_letrec_check_frame(FRAME_TYPE_CLOSURE, LET_BODY_EXPR,
-                                      num_params, inner, NULL, 
-                                      NULL);
-
-  val = letrec_check_expr(data->code, new_frame, scheme_false);
-  data->code = val;
+    inner = clos->frame;
+    
+    num_params = lam->num_params;
+    
+    new_frame = init_letrec_check_frame(FRAME_TYPE_CLOSURE, LET_BODY_EXPR,
+                                        num_params, inner, NULL, 
+                                        NULL, lam);
+    
+    val = letrec_check_expr(lam->body, new_frame, scheme_false);
+    lam->body = val;
+  } else {
+    letrec_check_expr(clos->expr, clos->frame, scheme_false);
+  }
 }
 
 static void clean_dead_deferred_expr(Scheme_Deferred_Expr *clos)
 {
-  Scheme_Closure_Data *data;
+  Scheme_Lambda *lam;
 
   /* We keep a global chain of all deferred expression. A deferred
      expression that is never forced is a function that is never
@@ -695,12 +651,14 @@ static void clean_dead_deferred_expr(Scheme_Deferred_Expr *clos)
                   "letrec_check_deferred_expr: clos is not a scheme_deferred_expr");
 
     if (!clos->done) {
-      data = (Scheme_Closure_Data *)clos->expr;
-      SCHEME_ASSERT(SAME_TYPE(SCHEME_TYPE(data), scheme_compiled_unclosed_procedure_type),
-                    "deferred expression does not contain a lambda");
+      if (SAME_TYPE(SCHEME_TYPE(clos->expr), scheme_ir_lambda_type)) {
+        lam = (Scheme_Lambda *)clos->expr;
+        SCHEME_ASSERT(SAME_TYPE(SCHEME_TYPE(lam), scheme_ir_lambda_type),
+                      "deferred expression does not contain a lambda");
       
-      /* Since this deferral was never done, it's dead code. */
-      data->code = scheme_void;
+        /* Since this deferral was never done, it's dead code. */
+        lam->body = scheme_void;
+      }
 
       clos->done = 1;
     }
@@ -727,14 +685,14 @@ static void process_deferred_bindings(Letrec_Check_Frame *frame, int position)
 static Scheme_Object *letrec_check_lets(Scheme_Object *o, Letrec_Check_Frame *old_frame, Scheme_Object *pos)
 {
   Letrec_Check_Frame *frame, *body_frame;
-  Scheme_Compiled_Let_Value *clv;
+  Scheme_IR_Let_Value *irlv;
   Scheme_Object *body, *val;
   int i, j, k;
 
   /* gets the information out of our header about the number of
      total clauses, the number of total bindings, and whether or not
      this let is recursive */
-  Scheme_Let_Header *head = (Scheme_Let_Header *)o;
+  Scheme_IR_Let_Header *head = (Scheme_IR_Let_Header *)o;
 
   /* number of clauses in the let */
   int num_clauses = head->num_clauses;
@@ -751,8 +709,6 @@ static Scheme_Object *letrec_check_lets(Scheme_Object *o, Letrec_Check_Frame *ol
 
   if (header_flags & SCHEME_LET_RECURSIVE)
     frame_type = FRAME_TYPE_LETREC; 
-  else if (header_flags & SCHEME_LET_STAR)
-    frame_type = FRAME_TYPE_LETSTAR; 
   else
     frame_type = FRAME_TYPE_LET; 
 
@@ -761,25 +717,19 @@ static Scheme_Object *letrec_check_lets(Scheme_Object *o, Letrec_Check_Frame *ol
      resolving local-variable offsets */
   frame = init_letrec_check_frame(frame_type, LET_RHS_EXPR,
                                   count, old_frame, NULL,
-                                  head);
+                                  head, NULL);
 
   body = head->body;
 
-  if (frame_type == FRAME_TYPE_LETREC)
-    k = head->count;
-  else
-    k = 0;
+  k = 0;
 
   /* loops through every right hand side */
-  clv = NULL;
+  irlv = NULL;
   for (i = num_clauses; i--;) {
-    clv = (Scheme_Compiled_Let_Value *)body;
+    irlv = (Scheme_IR_Let_Value *)body;
 
-    if (frame_type == FRAME_TYPE_LETREC)
-      k -= clv->count;
-
-    if (clv->count == 0) {
-      val = letrec_check_expr(clv->value, frame,
+    if (irlv->count == 0) {
+      val = letrec_check_expr(irlv->value, frame,
                               /* deferred closures get attached to no variables,
                                  which is sensible because the closure will not
                                  be reachable: */
@@ -787,7 +737,7 @@ static Scheme_Object *letrec_check_lets(Scheme_Object *o, Letrec_Check_Frame *ol
     } else {
       Scheme_Object *new_pos;
 
-      if (clv->count == 1) {
+      if (irlv->count == 1) {
         /* any deferred closure on the right-hand side gets attached to the
            variable on the left-hand side: */
         new_pos = scheme_make_integer(k);
@@ -798,26 +748,25 @@ static Scheme_Object *letrec_check_lets(Scheme_Object *o, Letrec_Check_Frame *ol
            variables in that case */
         int sub;
         new_pos = scheme_null;
-        for (sub = clv->count; sub--; ) {
+        for (sub = irlv->count; sub--; ) {
           new_pos = scheme_make_pair(scheme_make_integer(k+sub), new_pos);
         }
       }
 
-      val = letrec_check_expr(clv->value, frame, new_pos);
+      val = letrec_check_expr(irlv->value, frame, new_pos);
     }
 
-    if (frame_type != FRAME_TYPE_LETREC)
-      k += clv->count;
-
     if (frame_type == FRAME_TYPE_LETREC) {
-      for (j = 0; j < clv->count; j++) {
+      for (j = 0; j < irlv->count; j++) {
         frame->ref[j + k] |= LET_READY;
       }
     }
-        
-    clv->value = val;
 
-    body = clv->body;
+    k += irlv->count;
+        
+    irlv->value = val;
+
+    body = irlv->body;
   }
 
   /* the body variant of the frame shares the `ref`, etc., arrays with
@@ -826,7 +775,7 @@ static Scheme_Object *letrec_check_lets(Scheme_Object *o, Letrec_Check_Frame *ol
      information) */
   body_frame = init_letrec_check_frame(frame_type, LET_BODY_EXPR,
                                        count, old_frame, frame,
-                                       head);
+                                       head, NULL);
   
   val = letrec_check_expr(body, body_frame, pos);
 
@@ -834,7 +783,7 @@ static Scheme_Object *letrec_check_lets(Scheme_Object *o, Letrec_Check_Frame *ol
      let had bindings, otherwise, the let header should point to the
      new body */
   if (num_clauses > 0)
-    clv->body = val;
+    irlv->body = val;
   else
     head->body = val;
 
@@ -845,27 +794,24 @@ static Scheme_Object *letrec_check_lets(Scheme_Object *o, Letrec_Check_Frame *ol
 
 /* note to future self: the length of define_values is sometimes 1,
    and you definitely don't want to look inside if that's the case */
-static Scheme_Object *letrec_check_define_values(Scheme_Object *data, Letrec_Check_Frame *frame, Scheme_Object *pos)
+static Scheme_Object *letrec_check_define_values(Scheme_Object *lam, Letrec_Check_Frame *frame, Scheme_Object *pos)
 {
-  if (SCHEME_VEC_SIZE(data) <= 1)
-    return data;
+  if (SCHEME_VEC_SIZE(lam) <= 1)
+    return lam;
   else {
-    Scheme_Object *vars = SCHEME_VEC_ELS(data)[0];
-    Scheme_Object *val = SCHEME_VEC_ELS(data)[1];
-    SCHEME_ASSERT(SCHEME_PAIRP(vars) || SCHEME_NULLP(vars),
-                  "letrec_check_define_values: processing resolved code");
+    Scheme_Object *val = SCHEME_VEC_ELS(lam)[0];
 
     val = letrec_check_expr(val, frame, pos);
 
-    SCHEME_VEC_ELS(data)[1] = val;
+    SCHEME_VEC_ELS(lam)[0] = val;
   }
     
-  return data;
+  return lam;
 }
 
-static Scheme_Object *letrec_check_ref(Scheme_Object *data, Letrec_Check_Frame *frame, Wrapped_Lhs *lhs)
+static Scheme_Object *letrec_check_ref(Scheme_Object *lam, Letrec_Check_Frame *frame, Wrapped_Lhs *lhs)
 {
-  return data;
+  return lam;
 }
 
 static Scheme_Object *letrec_check_set(Scheme_Object *o, Letrec_Check_Frame *frame, Scheme_Object *pos)
@@ -883,12 +829,12 @@ static Scheme_Object *letrec_check_set(Scheme_Object *o, Letrec_Check_Frame *fra
   val = letrec_check_expr(val, frame, rhs_pos);
   sb->val = val;
 
-  if (SAME_TYPE(SCHEME_TYPE(sb->var), scheme_local_type)) {
+  if (SAME_TYPE(SCHEME_TYPE(sb->var), scheme_ir_local_type)) {
     /* We may need to insert a definedness check before the assignment */
     Letrec_Check_Frame *in_frame;
-    int position = SCHEME_LOCAL_POS(sb->var);
+    int position;
 
-    in_frame = get_relative_frame(&position, frame);
+    in_frame = get_relative_frame(&position, (Scheme_IR_Local *)sb->var);
     
     if (in_frame->ref
         && !(in_frame->ref[position] & LET_READY)) {
@@ -896,55 +842,68 @@ static Scheme_Object *letrec_check_set(Scheme_Object *o, Letrec_Check_Frame *fra
       Scheme_App3_Rec *app3;
       Scheme_Object *name;
       Scheme_Sequence *seq;
-      
-      name = record_checked((Scheme_Local *)sb->var, frame);
-      
+      Scheme_IR_Let_Header *head;
+      Scheme_IR_Let_Value *lv;
+      Scheme_IR_Local *var, **vars;
+
+      /* Change
+           (set! <id> <rhs>)
+
+         to
+
+           (let ([<tmp> <rhs>])
+             (begin
+               <check>
+               (set! <id> <tmp>)))
+      */
+
+      name = record_checked((Scheme_IR_Local *)sb->var, frame);
+
       app3 = MALLOC_ONE_TAGGED(Scheme_App3_Rec);
       app3->iso.so.type = scheme_application3_type;
       app3->rator = scheme_check_assign_not_undefined_proc;
       app3->rand1 = sb->var;
       app3->rand2 = name;
-      
+
       seq = scheme_malloc_sequence(2);
       seq->so.type = scheme_sequence_type;
       seq->count = 2;
       seq->array[0] = (Scheme_Object *)app3;
       seq->array[1] = (Scheme_Object *)sb;
 
-      return (Scheme_Object *)seq;
+      if (SCHEME_TYPE(sb->val) > _scheme_ir_values_types_) {
+        /* obviously simple enough to not need a `let` wrapper */
+        return (Scheme_Object *)seq;
+      } else {
+        var = scheme_make_ir_local(scheme_intern_symbol("tmp"));
+        vars = MALLOC_N(Scheme_IR_Local*, 1);
+        vars[0] = var;
+        var->use_count = 1;
+        var->non_app_count = 1;
+
+        lv = MALLOC_ONE_TAGGED(Scheme_IR_Let_Value);
+        lv->iso.so.type = scheme_ir_let_value_type;
+        lv->count = 1;
+        lv->value = sb->val;
+        lv->body = (Scheme_Object *)seq;
+        lv->vars = vars;
+
+        head = MALLOC_ONE_TAGGED(Scheme_IR_Let_Header);
+        head->iso.so.type = scheme_ir_let_header_type;
+        head->count = 1;
+        head->num_clauses = 1;
+        head->body = (Scheme_Object *)lv;
+
+        sb->val = (Scheme_Object *)var;
+
+        return (Scheme_Object *)head;
+      }
     }
   }
 
   return o;
 }
 
-static Scheme_Object *letrec_check_define_syntaxes(Scheme_Object *data, Letrec_Check_Frame *frame, Scheme_Object *pos)
-{
-  Scheme_Object *val;
-  val = SCHEME_VEC_ELS(data)[3];
-
-  val = letrec_check_expr(val, frame, pos);
-  SCHEME_VEC_ELS(data)[3] = val;
-
-  return data;
-}
-
-static Scheme_Object *letrec_check_begin_for_syntax(Scheme_Object *data, Letrec_Check_Frame *frame, Scheme_Object *pos)
-{
-  Scheme_Object *l, *a, *val;
-    
-  l = SCHEME_VEC_ELS(data)[2];
-    
-  while (!SCHEME_NULLP(l)) {
-    a = SCHEME_CAR(l);
-    val = letrec_check_expr(a, frame, pos);
-    SCHEME_CAR(l) = val;
-    l = SCHEME_CDR(l);
-  }
-    
-  return data;
-}
-
 static Scheme_Object *letrec_check_case_lambda(Scheme_Object *o, Letrec_Check_Frame *frame, Scheme_Object *pos)
 {
   Scheme_Case_Lambda *cl;
@@ -979,68 +938,59 @@ static Scheme_Object *letrec_check_begin0(Scheme_Object *o, Letrec_Check_Frame *
   return o;
 }
 
-static Scheme_Object *letrec_check_apply_values(Scheme_Object *data, Letrec_Check_Frame *frame, Scheme_Object *pos)
+static Scheme_Object *letrec_check_apply_values(Scheme_Object *lam, Letrec_Check_Frame *frame, Scheme_Object *pos)
 {
   Scheme_Object *f, *e;
     
-  f = SCHEME_PTR1_VAL(data);
-  e = SCHEME_PTR2_VAL(data);
+  f = SCHEME_PTR1_VAL(lam);
+  e = SCHEME_PTR2_VAL(lam);
     
   f = letrec_check_expr(f, frame, pos);
   e = letrec_check_expr(e, frame, pos);
 
-  SCHEME_PTR1_VAL(data) = f;
-  SCHEME_PTR2_VAL(data) = e;
+  SCHEME_PTR1_VAL(lam) = f;
+  SCHEME_PTR2_VAL(lam) = e;
     
-  return data;
+  return lam;
 }
 
-static Scheme_Object *letrec_check_module(Scheme_Object *o, Letrec_Check_Frame *frame, Scheme_Object *pos)
+static Scheme_Object *letrec_check_k(void)
 {
-  int i, cnt;
-  Scheme_Module *m;
-  Scheme_Object *val;
-  m = (Scheme_Module *)o;
+  Scheme_Thread *p = scheme_current_thread;
+  Scheme_Object *expr = (Scheme_Object *)p->ku.k.p1;
+  Letrec_Check_Frame *frame = (Letrec_Check_Frame *)p->ku.k.p2;
+  Scheme_Object *pos = (Scheme_Object *)p->ku.k.p3;
 
-  if (!m->comp_prefix) {
-    /* already resolved */
-    return (Scheme_Object *)m;
-  }
-    
-  cnt = SCHEME_VEC_SIZE(m->bodies[0]);
-  for(i = 0; i < cnt; i++) {
-    val = SCHEME_VEC_ELS(m->bodies[0])[i];
-    val = letrec_check_expr(val, frame, pos);
-    SCHEME_VEC_ELS(m->bodies[0])[i] = val;
-  }
+  p->ku.k.p1 = NULL;
+  p->ku.k.p2 = NULL;
+  p->ku.k.p3 = NULL;
 
-  {
-    /* check submodules */
-    int k;
-    Scheme_Object *p;
-    for (k = 0; k < 2; k++) {
-      p = (k ? m->post_submodules : m->pre_submodules);
-      if (p) {
-        while (!SCHEME_NULLP(p)) {
-          letrec_check_expr(SCHEME_CAR(p), frame, pos);
-          p = SCHEME_CDR(p);
-        }
-      }
-    }
-  }
-
-  return o;
+  return letrec_check_expr(expr, frame, pos);
 }
 
 static Scheme_Object *letrec_check_expr(Scheme_Object *expr, Letrec_Check_Frame *frame, Scheme_Object *pos)
 {
   int type;
+
+#ifdef DO_STACK_CHECK
+# include "mzstkchk.h"
+  {
+    Scheme_Thread *p = scheme_current_thread;
+
+    p->ku.k.p1 = (void *)expr;
+    p->ku.k.p2 = (void *)frame;
+    p->ku.k.p3 = (void *)pos;
+
+    return scheme_handle_stack_overflow(letrec_check_k);
+  }
+#endif
+
   type = SCHEME_TYPE(expr);
 
   SCHEME_USE_FUEL(1);
 
   switch (type) {
-  case scheme_local_type:
+  case scheme_ir_local_type:
     return letrec_check_local(expr, frame, pos);
   case scheme_application_type:
     return letrec_check_application(expr, frame, pos);
@@ -1049,22 +999,18 @@ static Scheme_Object *letrec_check_expr(Scheme_Object *expr, Letrec_Check_Frame
   case scheme_application3_type:
     return letrec_check_application3(expr, frame, pos);
   case scheme_sequence_type:
-  case scheme_splice_sequence_type:
     return letrec_check_sequence(expr, frame, pos);
   case scheme_branch_type:
     return letrec_check_branch(expr, frame, pos);
   case scheme_with_cont_mark_type:
     return letrec_check_wcm(expr, frame, pos);
-  case scheme_compiled_unclosed_procedure_type:
-    return letrec_check_closure_compilation(expr, frame, pos);
-  case scheme_compiled_let_void_type:
+  case scheme_ir_lambda_type:
+    return letrec_check_lambda(expr, frame, pos);
+  case scheme_ir_let_header_type:
     return letrec_check_lets(expr, frame, pos);
-  case scheme_compiled_toplevel_type: /* var ref to a top level */
-    return expr;
-  case scheme_compiled_quote_syntax_type:
+  case scheme_ir_toplevel_type: /* var ref to a top level */
     return expr;
   case scheme_variable_type:
-  case scheme_module_variable_type:
     scheme_signal_error("got top-level in wrong place");
     return 0;
   case scheme_define_values_type:
@@ -1073,10 +1019,6 @@ static Scheme_Object *letrec_check_expr(Scheme_Object *expr, Letrec_Check_Frame
     return letrec_check_ref(expr, frame, pos);
   case scheme_set_bang_type:
     return letrec_check_set(expr, frame, pos);
-  case scheme_define_syntaxes_type:
-    return letrec_check_define_syntaxes(expr, frame, pos);
-  case scheme_begin_for_syntax_type:
-    return letrec_check_begin_for_syntax(expr, frame, pos);
   case scheme_case_lambda_sequence_type:
     return letrec_check_case_lambda(expr, frame, pos);
   case scheme_begin0_sequence_type:
@@ -1086,24 +1028,21 @@ static Scheme_Object *letrec_check_expr(Scheme_Object *expr, Letrec_Check_Frame
   case scheme_with_immed_mark_type:
     scheme_signal_error("internal error: with-immediate-mark not expected before optimization");
     return NULL;
-  case scheme_require_form_type:
-    return expr;
-  case scheme_module_type:
-    return letrec_check_module(expr, frame, pos);
   default:
     return expr;
   }
 }
 
-Scheme_Object *scheme_letrec_check_expr(Scheme_Object *expr)
+Scheme_Linklet *scheme_letrec_check_linklet(Scheme_Linklet *linklet)
 {
+  int i, cnt;
   Scheme_Object *val;
   Scheme_Object *init_pos = scheme_false;
   Letrec_Check_Frame *frame;
 
   frame = init_letrec_check_frame(FRAME_TYPE_TOP, LET_BODY_EXPR,
                                   0, NULL, NULL,
-                                  NULL);
+                                  NULL, NULL);
 
   /* (Grep "EXPL-3" for information): The `pos` argument, starting
      here as `init_pos`, represents whether we're in a non-application
@@ -1113,11 +1052,16 @@ Scheme_Object *scheme_letrec_check_expr(Scheme_Object *expr)
      positions. We use a list of numbers for the RHS of a
      `let[rec]-values` form with multiple variables. */
 
-  val = letrec_check_expr(expr, NULL, init_pos);
+  cnt = SCHEME_VEC_SIZE(linklet->bodies);
+  for(i = 0; i < cnt; i++) {
+    val = SCHEME_VEC_ELS(linklet->bodies)[i];
+    val = letrec_check_expr(val, frame, init_pos);
+    SCHEME_VEC_ELS(linklet->bodies)[i] = val;
+  }
 
   clean_dead_deferred_expr(*frame->deferred_chain);
 
-  return val;
+  return linklet;
 }
 
 /*========================================================================*/
