@@ -13,19 +13,6 @@ variable_obj {
   gcBYTES_TO_WORDS(sizeof(Scheme_Bucket_With_Home));
 }
 
-module_var {
- mark:
-  Module_Variable *mv = (Module_Variable *)p;
-
-  gcMARK2(mv->modidx, gc);
-  gcMARK2(mv->sym, gc);
-  gcMARK2(mv->insp, gc);
-  gcMARK2(mv->shape, gc);
-
- size:
-  gcBYTES_TO_WORDS(sizeof(Module_Variable));
-}
-
 bucket_obj {
  mark:
   Scheme_Bucket *b = (Scheme_Bucket *)p;
@@ -49,10 +36,11 @@ toplevel_obj {
   gcBYTES_TO_WORDS(sizeof(Scheme_Toplevel));
 }
 
-quotesyntax_obj {
+static_toplevel_obj {
  mark:
+  gcMARK2(SCHEME_STATIC_TOPLEVEL_PREFIX(p), gc);
  size:
-  gcBYTES_TO_WORDS(sizeof(Scheme_Quote_Syntax));
+  gcBYTES_TO_WORDS(sizeof(Scheme_Toplevel));
 }
 
 cpointer_obj {
@@ -234,6 +222,9 @@ ir_local {
 
   gcMARK2(var->name, gc);
   switch (var->mode) {
+  case SCHEME_VAR_MODE_COMPILE:
+    gcMARK2(var->compile.use_box, gc);
+    break;
   case SCHEME_VAR_MODE_LETREC_CHECK:
     gcMARK2(var->letrec_check.frame, gc);
     break;
@@ -252,6 +243,12 @@ ir_local {
   gcBYTES_TO_WORDS(sizeof(Scheme_IR_Local));
 }
 
+ir_toplevel {
+ mark:
+ size:
+  gcBYTES_TO_WORDS(sizeof(Scheme_IR_Toplevel));
+}
+
 ir_let_value {
  mark:
   Scheme_IR_Let_Value *c = (Scheme_IR_Let_Value *)p;
@@ -690,13 +687,6 @@ output_port {
   gcBYTES_TO_WORDS(sizeof(Scheme_Output_Port));
 }
 
-
-syntax_compiler {
- mark:
- size:
-  gcBYTES_TO_WORDS(sizeof(Scheme_Simple_Object));
-}
-
 thread_val {
  mark:
   Scheme_Thread *pr = (Scheme_Thread *)p;
@@ -752,14 +742,6 @@ thread_val {
   gcMARK2(pr->return_marks_to, gc);
   gcMARK2(pr->returned_marks, gc);
   
-  gcMARK2(pr->current_local_env, gc);
-  gcMARK2(pr->current_local_scope, gc);
-  gcMARK2(pr->current_local_use_scope, gc);
-  gcMARK2(pr->current_local_name, gc);
-  gcMARK2(pr->current_local_modidx, gc);
-  gcMARK2(pr->current_local_menv, gc);
-  gcMARK2(pr->current_local_bindings, gc);
-
   gcMARK2(pr->current_mt, gc);
 
   gcMARK2(pr->constant_folding, gc);
@@ -828,8 +810,8 @@ runstack_val {
   intptr_t *s = (intptr_t *)p;
  mark:
   void **a, **b;
-  a = (void **)s + 5 + s[2];
-  b = (void **)s + 5 + s[3];
+  a = (void **)s + RUNSTACK_HEADER_FIELDS + s[2];
+  b = (void **)s + RUNSTACK_HEADER_FIELDS + s[3];
   while (a < b) {
     gcMARK2(*a, gc);
     a++;
@@ -838,14 +820,14 @@ runstack_val {
   START_MARK_ONLY;
   /* Zero out the part that we didn't mark, in case it becomes
      live later. */
-  a = (void **)s + 5;
-  b = (void **)s + 5 + s[2];
+  a = (void **)s + RUNSTACK_HEADER_FIELDS;
+  b = (void **)s + RUNSTACK_HEADER_FIELDS + s[2];
   while (a < b) {
     *a = RUNSTACK_ZERO_VAL;
     a++;
   }
-  a = (void **)s + 5 + s[3];
-  b = (void **)s + 5 + (s[1] - 5);
+  a = (void **)s + RUNSTACK_HEADER_FIELDS + s[3];
+  b = (void **)s + RUNSTACK_HEADER_FIELDS + (s[1] - RUNSTACK_HEADER_FIELDS);
   while (a < b) {
     *a = RUNSTACK_ZERO_VAL;
     a++;
@@ -860,8 +842,12 @@ prompt_val {
  mark: 
   Scheme_Prompt *pr = (Scheme_Prompt *)p;
   gcMARK2(pr->boundary_overflow_id, gc);
-  if (!GC_merely_accounting())
-    gcMARK2(pr->runstack_boundary_start, gc);
+  if (!GC_merely_accounting()) {
+    if (pr->is_barrier)
+      gcMARK2(pr->u.runstack_boundary_start_ref, gc);
+    else
+      gcMARK2(pr->u.runstack_boundary_start, gc);
+  }
   gcMARK2(pr->tag, gc);
   gcMARK2(pr->id, gc);
  size:
@@ -936,61 +922,27 @@ bucket_table_val {
   gcBYTES_TO_WORDS(sizeof(Scheme_Bucket_Table));
 }
 
-namespace_val {
+env_val {
  mark:
   Scheme_Env *e = (Scheme_Env *)p;
 
-  gcMARK2(e->module, gc);
-  gcMARK2(e->module_registry, gc);
-  gcMARK2(e->module_pre_registry, gc);
-  gcMARK2(e->guard_insp, gc);
-  gcMARK2(e->access_insp, gc);
-
-  gcMARK2(e->stx_context, gc);
-  gcMARK2(e->tmp_bind_scope, gc);
-
-  gcMARK2(e->syntax, gc);
-  gcMARK2(e->exp_env, gc);
-  gcMARK2(e->template_env, gc);
-  gcMARK2(e->label_env, gc);
-  gcMARK2(e->instance_env, gc);
-  gcMARK2(e->reader_env, gc);
-
-  gcMARK2(e->shadowed_syntax, gc);
-
-  gcMARK2(e->lift_key, gc);
-
-  gcMARK2(e->link_midx, gc);
-  gcMARK2(e->require_names, gc);
-  gcMARK2(e->et_require_names, gc);
-  gcMARK2(e->tt_require_names, gc);
-  gcMARK2(e->dt_require_names, gc);
-  gcMARK2(e->other_require_names, gc);
-  gcMARK2(e->running, gc);
-  gcMARK2(e->did_starts, gc);
-  gcMARK2(e->available_next[0], gc);
-  gcMARK2(e->available_next[1], gc);
-
-  gcMARK2(e->toplevel, gc);
-  gcMARK2(e->modchain, gc);
-
-  gcMARK2(e->modvars, gc);
-
-  gcMARK2(e->weak_self_link, gc);
-
-  gcMARK2(e->binding_names, gc);
-
+  gcMARK2(e->namespace, gc);
+  gcMARK2(e->instance, gc);
+  gcMARK2(e->protected, gc);
  size:
   gcBYTES_TO_WORDS(sizeof(Scheme_Env));
 }
 
-module_reg_val {
+startup_env_val {
  mark:
-  Scheme_Module_Registry *r = (Scheme_Module_Registry *)p;
-  gcMARK2(r->loaded, gc);
-  gcMARK2(r->exports, gc);
+  Scheme_Startup_Env *e = (Scheme_Startup_Env *)p;
+
+  gcMARK2(e->current_table, gc);
+  gcMARK2(e->primitive_tables, gc);
+  gcMARK2(e->all_primitives_table, gc);
+  gcMARK2(e->primitive_ids_table, gc);
  size:
-  gcBYTES_TO_WORDS(sizeof(Scheme_Module_Registry));
+  gcBYTES_TO_WORDS(sizeof(Scheme_Startup_Env));
 }
 
 random_state_val {
@@ -999,17 +951,6 @@ random_state_val {
   gcBYTES_TO_WORDS(sizeof(Scheme_Random_State));
 }
 
-compilation_top_val {
- mark:
-  Scheme_Compilation_Top *t = (Scheme_Compilation_Top *)p;
-  gcMARK2(t->code, gc);
-  gcMARK2(t->prefix, gc);
-  gcMARK2(t->binding_namess, gc);
-
- size:
-  gcBYTES_TO_WORDS(sizeof(Scheme_Compilation_Top));
-}
-
 prefix_val {
   Scheme_Prefix *pf = (Scheme_Prefix *)p;
  mark:
@@ -1019,34 +960,10 @@ prefix_val {
  size:
   gcBYTES_TO_WORDS((sizeof(Scheme_Prefix) 
 		    + ((pf->num_slots-mzFLEX_DELTA) * sizeof(Scheme_Object *))
-                    + ((((pf->num_slots - pf->num_stxes) + 31) / 32) 
+                    + ((((pf->num_slots + 31) / 32) 
                        * sizeof(int))));
 }
 
-resolve_prefix_val {
- mark:
-  Resolve_Prefix *rp = (Resolve_Prefix *)p;
-  gcMARK2(rp->toplevels, gc);
-  gcMARK2(rp->stxes, gc);
-  gcMARK2(rp->delay_info_rpair, gc);
-  gcMARK2(rp->src_insp_desc, gc);
-
- size:
-  gcBYTES_TO_WORDS(sizeof(Resolve_Prefix));
-}
-
-comp_prefix_val {
- mark:
-  Comp_Prefix *cp = (Comp_Prefix *)p;
-  gcMARK2(cp->toplevels, gc);
-  gcMARK2(cp->inline_variants, gc);
-  gcMARK2(cp->unbound, gc);
-  gcMARK2(cp->stxes, gc);
-
- size:
-  gcBYTES_TO_WORDS(sizeof(Comp_Prefix));
-}
-
 svector_val {
  mark:
   Scheme_Object *o = (Scheme_Object *)p;
@@ -1062,139 +979,39 @@ stx_val {
   Scheme_Stx *stx = (Scheme_Stx *)p;
   gcMARK2(stx->val, gc);
   gcMARK2(stx->srcloc, gc);
-  gcMARK2(stx->scopes, gc);
-  gcMARK2(stx->u.to_propagate, gc);
-  gcMARK2(stx->shifts, gc);
-  gcMARK2(stx->taints, gc);
   gcMARK2(stx->props, gc);
  size:
   gcBYTES_TO_WORDS(sizeof(Scheme_Stx));
 }
 
-stx_off_val {
+linklet_val {
  mark:
-  Scheme_Stx_Offset *o = (Scheme_Stx_Offset *)p;
-  gcMARK2(o->src, gc);
+  Scheme_Linklet *l = (Scheme_Linklet *)p;
+
+  gcMARK2(l->name, gc);
+  gcMARK2(l->importss, gc);
+  gcMARK2(l->import_shapes, gc);
+  gcMARK2(l->defns, gc);
+  gcMARK2(l->source_names, gc);
+  gcMARK2(l->bodies, gc);
+  gcMARK2(l->constants, gc);
+  gcMARK2(l->static_prefix, gc);
+  gcMARK2(l->native_lambdas, gc);
  size:
-  gcBYTES_TO_WORDS(sizeof(Scheme_Stx_Offset));
+  gcBYTES_TO_WORDS(sizeof(Scheme_Linklet));
 }
 
-module_val {
+instance_val {
  mark:
-  Scheme_Module *m = (Scheme_Module *)p;
+  Scheme_Instance *i = (Scheme_Instance *)p;
 
-  gcMARK2(m->phaseless, gc);
-
-  gcMARK2(m->code_key, gc);
-
-  gcMARK2(m->modname, gc);
-  gcMARK2(m->modsrc, gc);
-
-  gcMARK2(m->et_requires, gc);
-  gcMARK2(m->requires, gc);
-  gcMARK2(m->tt_requires, gc);
-  gcMARK2(m->dt_requires, gc);
-  gcMARK2(m->other_requires, gc);
-
-  gcMARK2(m->bodies, gc);
-
-  gcMARK2(m->me, gc);
-
-  gcMARK2(m->exp_infos, gc);
-
-  gcMARK2(m->self_modidx, gc);
-
-  gcMARK2(m->binding_names, gc);
-  gcMARK2(m->et_binding_names, gc);
-  gcMARK2(m->other_binding_names, gc);
-
-  gcMARK2(m->insp, gc);
-
-  gcMARK2(m->lang_info, gc);
-
-  gcMARK2(m->hints, gc);
-  gcMARK2(m->ii_src, gc);
-  gcMARK2(m->super_bxs_info, gc);
-  gcMARK2(m->sub_iidx_ptrs, gc);
-
-  gcMARK2(m->comp_prefix, gc);
-  gcMARK2(m->prefix, gc);
-  gcMARK2(m->dummy, gc);
-
-  gcMARK2(m->rn_stx, gc);
-
-  gcMARK2(m->submodule_path, gc);
-  gcMARK2(m->pre_submodules, gc);
-  gcMARK2(m->post_submodules, gc);
-  gcMARK2(m->pre_submodule_names, gc);
-  gcMARK2(m->supermodule, gc);
-  gcMARK2(m->submodule_ancestry, gc);
-
-  gcMARK2(m->primitive, gc);
+  gcMARK2(i->variables.a, gc);
+  gcMARK2(i->weak_self_link, gc);
+  gcMARK2(i->source_names, gc);
+  gcMARK2(i->name, gc);
+  gcMARK2(i->data, gc);
  size:
-  gcBYTES_TO_WORDS(sizeof(Scheme_Module));
-}
-
-exp_info_val {
- mark:
-  Scheme_Module_Export_Info *m = (Scheme_Module_Export_Info *)p;
-
-  gcMARK2(m->provide_protects, gc);
-  gcMARK2(m->indirect_provides, gc);
-
-  gcMARK2(m->indirect_syntax_provides, gc);
-
-  gcMARK2(m->accessible, gc);
- size:
-  gcBYTES_TO_WORDS(sizeof(Scheme_Module_Export_Info));
-}
-
-module_phase_exports_val {
- mark:
-  Scheme_Module_Phase_Exports *m = (Scheme_Module_Phase_Exports *)p;
-
-  gcMARK2(m->phase_index, gc);
-
-  gcMARK2(m->src_modidx, gc);
-
-  gcMARK2(m->provides, gc);
-  gcMARK2(m->provide_srcs, gc);
-  gcMARK2(m->provide_src_names, gc);
-  gcMARK2(m->provide_nominal_srcs, gc);
-  gcMARK2(m->provide_src_phases, gc);
-
-  gcMARK2(m->ht, gc);
-
- size:
-  gcBYTES_TO_WORDS(sizeof(Scheme_Module_Phase_Exports));
-}
-
-module_exports_val {
- mark:
-  Scheme_Module_Exports *m = (Scheme_Module_Exports *)p;
-
-  gcMARK2(m->rt, gc);
-  gcMARK2(m->et, gc);
-  gcMARK2(m->dt, gc);
-  gcMARK2(m->other_phases, gc);
-
-  gcMARK2(m->src_modidx, gc);
-  gcMARK2(m->modsrc, gc);
- size:
-  gcBYTES_TO_WORDS(sizeof(Scheme_Module_Exports));
-}
-
-modidx_val {
- mark:
-  Scheme_Modidx *modidx = (Scheme_Modidx *)p;
-
-  gcMARK2(modidx->path, gc);
-  gcMARK2(modidx->base, gc);
-  gcMARK2(modidx->resolved, gc);
-  gcMARK2(modidx->shift_cache, gc);
-  gcMARK2(modidx->cache_next, gc);
- size:
-  gcBYTES_TO_WORDS(sizeof(Scheme_Modidx));
+  gcBYTES_TO_WORDS(sizeof(Scheme_Instance));
 }
 
 guard_val {
@@ -1248,6 +1065,7 @@ mark_logger {
   gcMARK2(l->root_timestamp, gc);
   gcMARK2(l->syslog_level, gc);
   gcMARK2(l->stderr_level, gc);
+  gcMARK2(l->stdout_level, gc);
   gcMARK2(l->propagate_level, gc);
   gcMARK2(l->readers, gc);
  size:
@@ -1283,36 +1101,21 @@ END env;
 
 /**********************************************************************/
 
+START linklet;
+
+END linklet;
+
+/**********************************************************************/
+
 START compenv;
 
 mark_comp_env {
  mark:
   Scheme_Comp_Env *e = (Scheme_Comp_Env *)p;
 
-  gcMARK2(e->genv, gc);
-  gcMARK2(e->insp, gc);
-  gcMARK2(e->prefix, gc);
-  gcMARK2(e->next, gc);
-  gcMARK2(e->scopes, gc);
-  gcMARK2(e->value_name, gc);
-  gcMARK2(e->observer, gc);
-  gcMARK2(e->binders, gc);
-  gcMARK2(e->bindings, gc);
-  gcMARK2(e->vals, gc);
-  gcMARK2(e->shadower_deltas, gc);
   gcMARK2(e->vars, gc);
-  gcMARK2(e->dup_check, gc);
-  gcMARK2(e->intdef_name, gc);
-  gcMARK2(e->in_modidx, gc);
-  gcMARK2(e->skip_table, gc);
-  
-  gcMARK2(e->use, gc);
-  gcMARK2(e->lifts, gc);
-  gcMARK2(e->bindings, gc);
-
-  gcMARK2(e->binding_namess, gc);
-
-  gcMARK2(e->expand_result_adjust_arg, gc);
+  gcMARK2(e->value_name, gc);
+  gcMARK2(e->linklet, gc);
 
  size:
   gcBYTES_TO_WORDS(sizeof(Scheme_Comp_Env));
@@ -1328,12 +1131,16 @@ mark_resolve_info {
  mark:
   Resolve_Info *i = (Resolve_Info *)p;
   
-  gcMARK2(i->prefix, gc);
-  gcMARK2(i->stx_map, gc);
+  gcMARK2(i->linklet, gc);
   gcMARK2(i->tl_map, gc);
   gcMARK2(i->redirects, gc);
   gcMARK2(i->lifts, gc);
+  gcMARK2(i->top, gc);
   gcMARK2(i->next, gc);
+  gcMARK2(i->toplevel_starts, gc);
+  gcMARK2(i->toplevel_deltas, gc);
+  gcMARK2(i->toplevel_defns, gc);
+  gcMARK2(i->static_mode, gc);
 
  size:
   gcBYTES_TO_WORDS(sizeof(Resolve_Info));
@@ -1344,16 +1151,10 @@ mark_unresolve_info {
   Unresolve_Info *i = (Unresolve_Info *)p;
   
   gcMARK2(i->vars, gc);
-  gcMARK2(i->prefix, gc);
+  gcMARK2(i->linklet, gc);
+  gcMARK2(i->linklet_key, gc);
+  gcMARK2(i->opt_info, gc);
   gcMARK2(i->closures, gc);
-  gcMARK2(i->module, gc);
-  gcMARK2(i->comp_prefix, gc);
-  gcMARK2(i->new_toplevels, gc);
-  gcMARK2(i->from_modidx, gc);
-  gcMARK2(i->to_modidx, gc);
-  gcMARK2(i->opt_env, gc);
-  gcMARK2(i->opt_insp, gc);
-  gcMARK2(i->inline_variants, gc);
   gcMARK2(i->toplevels, gc);
   gcMARK2(i->definitions, gc);
   gcMARK2(i->ref_lifts, gc);
@@ -1422,10 +1223,9 @@ mark_optimize_info {
   Optimize_Info *i = (Optimize_Info *)p;
   
   gcMARK2(i->next, gc);
-  gcMARK2(i->consts, gc);
-  gcMARK2(i->cp, gc);
-  gcMARK2(i->env, gc);
-  gcMARK2(i->insp, gc);
+  gcMARK2(i->linklet, gc);
+  gcMARK2(i->cross, gc);
+  gcMARK2(i->imports_used, gc);
   gcMARK2(i->top_level_consts, gc);
   gcMARK2(i->transitive_use_var, gc);
   gcMARK2(i->context, gc);
@@ -1643,20 +1443,6 @@ END place;
 
 START portfun;
 
-mark_load_handler_data {
- mark:
-  LoadHandlerData *d = (LoadHandlerData *)p;
-    
-  gcMARK2(d->config, gc);
-  gcMARK2(d->port, gc);
-  gcMARK2(d->p, gc);
-  gcMARK2(d->stxsrc, gc);
-  gcMARK2(d->expected_module, gc);
-  
- size:
-  gcBYTES_TO_WORDS(sizeof(LoadHandlerData));
-}
-
 mark_indexed_string {
  mark:
   Scheme_Indexed_String *is = (Scheme_Indexed_String *)p;
@@ -1730,52 +1516,23 @@ mark_output_file {
   gcBYTES_TO_WORDS(sizeof(Scheme_Output_File));
 }
 
-#ifdef MZ_FDS
 mark_input_fd {
  mark:
   Scheme_FD *fd = (Scheme_FD *)p;
 
   gcMARK2(fd->buffer, gc);
-  gcMARK2(fd->refcount, gc);
+  /* fd->refcount is malloc()ed */
   gcMARK2(fd->flush_handle, gc);
+  gcMARK2(fd->bufwidths, gc);
 
  size:
   gcBYTES_TO_WORDS(sizeof(Scheme_FD));
 }
-#endif
-
-#if defined(UNIX_PROCESSES) && !(defined(MZ_USE_PLACES) && defined(MZ_PRECISE_GC))
-mark_system_child {
- mark:
-  System_Child *sc = (System_Child *)p;
-
-  gcMARK2(sc->next, gc);
-
- size:
-  gcBYTES_TO_WORDS(sizeof(System_Child));
-}
-#endif
-
-#ifdef USE_OSKIT_CONSOLE
-mark_oskit_console_input {
- mark:
-  osk_console_input *c = (osk_console_input *)p;
-    
-  gcMARK2(c->buffer, gc);
-  gcMARK2(c->next, gc);
-
- size:
-  gcBYTES_TO_WORDS(sizeof(osk_console_input));
-}
-#endif
 
 mark_subprocess {
  mark:
-#ifndef WINDOWS_PROCESSES
   Scheme_Subprocess *sp = (Scheme_Subprocess *)p;
-  gcMARK2(sp->handle, gc);
   gcMARK2(sp->mref, gc);
-#endif
  size:
   gcBYTES_TO_WORDS(sizeof(Scheme_Subprocess));
 }
@@ -1793,7 +1550,6 @@ mark_read_write_evt {
 mark_filesystem_change_evt {
  mark:
   Scheme_Filesystem_Change_Evt *fc = (Scheme_Filesystem_Change_Evt *)p;
-  gcMARK2(fc->sema, gc);
   gcMARK2(fc->mref, gc);
  size:
   gcBYTES_TO_WORDS(sizeof(Scheme_Filesystem_Change_Evt));
@@ -1823,17 +1579,10 @@ mark_marshal_tables {
   gcMARK2(mt->symtab, gc);
   gcMARK2(mt->st_refs, gc);
   gcMARK2(mt->st_ref_stack, gc);
-  gcMARK2(mt->reachable_scopes, gc);
-  gcMARK2(mt->reachable_scope_stack, gc);
-  gcMARK2(mt->pending_reachable_ids, gc);
-  gcMARK2(mt->conditionally_reachable_scopes, gc);
   gcMARK2(mt->intern_map, gc);
-  gcMARK2(mt->identity_map, gc);
-  gcMARK2(mt->top_map, gc);
   gcMARK2(mt->key_map, gc);
   gcMARK2(mt->delay_map, gc);
   gcMARK2(mt->cdata_map, gc);
-  gcMARK2(mt->rn_saved, gc);
   gcMARK2(mt->shared_offsets, gc);
   gcMARK2(mt->path_cache, gc);
   gcMARK2(mt->sorted_keys, gc);
@@ -1847,16 +1596,12 @@ END print;
 
 START network;
 
-#ifdef USE_TCP
 mark_listener {
   listener_t *l = (listener_t *)p;
 
  mark:
 
   gcMARK2(l->mref, gc);
-# ifdef HAVE_POLL_SYSCALL
-  gcMARK2(l->pfd, gc);
-# endif
 
  size:
   gcBYTES_TO_WORDS(sizeof(listener_t) + ((l->count - mzFLEX_DELTA) * sizeof(tcp_t)));
@@ -1873,7 +1618,6 @@ mark_tcp {
   gcBYTES_TO_WORDS(sizeof(Scheme_Tcp));
 }
 
-# ifdef UDP_IS_SUPPORTED
 mark_udp {
  mark:
   Scheme_UDP *udp = (Scheme_UDP *)p;
@@ -1891,14 +1635,10 @@ mark_udp_evt {
 
   gcMARK2(uw->udp, gc);
   gcMARK2(uw->str, gc);
-  gcMARK2(uw->dest_addrs, gc);
-  gcMARK2(uw->dest_addr_lens, gc);
 
  size:
   gcBYTES_TO_WORDS(sizeof(Scheme_UDP_Evt));
 }
-# endif
-#endif
 
 END network;
 
@@ -2306,8 +2046,6 @@ mark_cport {
   gcMARK2(cp->symtab, gc);
   gcMARK2(cp->symtab_entries, gc);
   gcMARK2(cp->relto, gc);
-  gcMARK2(cp->magic_sym, gc);
-  gcMARK2(cp->magic_val, gc);
   gcMARK2(cp->shared_offsets, gc);
   gcMARK2(cp->delay_info, gc);
   gcMARK2(cp->symtab_refs, gc);
@@ -2315,25 +2053,12 @@ mark_cport {
   gcBYTES_TO_WORDS(sizeof(CPort));
 }
 
-mark_readtable {
- mark:
-  Readtable *t = (Readtable *)p;
-  gcMARK2(t->mapping, gc);
-  gcMARK2(t->fast_mapping, gc);
-  gcMARK2(t->symbol_parser, gc);
-  gcMARK2(t->names, gc);
- size:
-  gcBYTES_TO_WORDS(sizeof(Readtable));
-}
-
 mark_read_params {
  mark:
   ReadParams *rp = (ReadParams *)p;
-  gcMARK2(rp->table, gc);
-  gcMARK2(rp->magic_sym, gc);
-  gcMARK2(rp->magic_val, gc);
   gcMARK2(rp->delay_load_info, gc);
   gcMARK2(rp->read_relative_path, gc);
+  gcMARK2(rp->graph_ht, gc);
  size:
   gcBYTES_TO_WORDS(sizeof(ReadParams));
 }
@@ -2357,10 +2082,6 @@ mark_delay_load {
 mark_unmarshal_tables {
  mark:
   Scheme_Unmarshal_Tables *ut = (Scheme_Unmarshal_Tables *)p;
-  gcMARK2(ut->rns, gc);
-  gcMARK2(ut->current_rns, gc);
-  gcMARK2(ut->multi_scope_pairs, gc);
-  gcMARK2(ut->current_multi_scope_pairs, gc);
   gcMARK2(ut->rp, gc);
   gcMARK2(ut->decoded, gc);
  size:
@@ -2438,40 +2159,6 @@ mark_srcloc {
   gcBYTES_TO_WORDS(sizeof(Scheme_Stx_Srcloc));
 }
 
-mark_scope {
-  Scheme_Scope *m = (Scheme_Scope *)p;
-  int for_multi = SCHEME_SCOPE_HAS_OWNER(m);
- mark:
-  gcMARK2(m->bindings, gc);
-  if (for_multi) {
-    gcMARK2(((Scheme_Scope_With_Owner *)m)->owner_multi_scope, gc);
-    gcMARK2(((Scheme_Scope_With_Owner *)m)->phase, gc);
-  }
- size:
- (for_multi
-  ? gcBYTES_TO_WORDS(sizeof(Scheme_Scope_With_Owner))
-  : gcBYTES_TO_WORDS(sizeof(Scheme_Scope)));
-}
-
-mark_scope_table {
- mark:
-  Scheme_Scope_Table *m = (Scheme_Scope_Table *)p;
-  gcMARK2(m->simple_scopes, gc);
-  gcMARK2(m->multi_scopes, gc);
- size:
-  gcBYTES_TO_WORDS(sizeof(Scheme_Scope_Table));
-}
-
-mark_propagate_table {
- mark:
-  Scheme_Propagate_Table *m = (Scheme_Propagate_Table *)p;
-  mark_scope_table_MARK(&m->st, gc);
-  gcMARK2(m->prev, gc);
-  gcMARK2(m->phase_shift, gc);
- size:
-  gcBYTES_TO_WORDS(sizeof(Scheme_Propagate_Table));
-}
-
 END syntax;
 
 /**********************************************************************/
